#ifndef __TEXTURE_H
#define __TEXTURE_H

#include <stdio.h>
#include <windows.h>
#include <commctrl.h>
#include <windowsx.h>
#include <gl/gl.h>
#include <gl/glu.h>
#include <gl/glext.h>
#include "logging.h"
#include "3dmath.h"

#define BITMAP_ID 0x4D42		// the universal bitmap ID

GLuint base;
bool multitex;
PFNGLMULTITEXCOORD2FARBPROC glMultiTexCoord2fARB = NULL;
PFNGLACTIVETEXTUREARBPROC   glActiveTextureARB = NULL;
PFNGLCLIENTACTIVETEXTUREARBPROC glClientActiveTextureARB = NULL;

int maxTextureUnits = 0;


//**********************************************************
//Initializes multitexturing and checks the amount of texture units
//**********************************************************

void MultiTexturing(){

     multitex=strstr((char*)glGetString(GL_EXTENSIONS),"GL_ARB_multitexture");
     

     if(multitex){
                  log_txt("Multitexturing supported!");
                  glGetIntegerv(GL_MAX_TEXTURE_UNITS_ARB, &maxTextureUnits);
                  log_txt("%i Texture Units",maxTextureUnits);

                  // retrieve addresses of multitexturing functions
                  glMultiTexCoord2fARB = (PFNGLMULTITEXCOORD2FARBPROC) wglGetProcAddress("glMultiTexCoord2fARB");
                  glActiveTextureARB = (PFNGLACTIVETEXTUREARBPROC) wglGetProcAddress("glActiveTextureARB");
                  glClientActiveTextureARB = (PFNGLCLIENTACTIVETEXTUREARBPROC) wglGetProcAddress("glClientActiveTextureARB");
     }
     else log_txt("No multitexturing support!");


}


enum texTypes_t
{
    BMP,
	TGA
};

typedef struct 
{
	// C Compatible version: enum texTypes textureType;
	// We use C++ only version
	texTypes_t textureType;
    bool alpha;
    
    char name[50];
    
	GLuint width;					// width of texture
	GLuint height;					// height of texture
    GLuint bpp;  //BitPerPixel
	long int scaledWidth;
	long int scaledHeight;

	GLuint texID;			// the texture object id of this texture
	GLubyte *data;		// the texture data
	
} texture_t;


// LoadBitmapFile
// desc: Returns a pointer to the bitmap image of the bitmap specified
//       by filename. Also returns the bitmap header information.
//		 No support for 8-bit bitmaps.
unsigned char *LoadOnlyBitmapFile(char *filename, BITMAPINFOHEADER *bitmapInfoHeader)
{
  FILE *filePtr;                        // the file pointer
  BITMAPFILEHEADER  bitmapFileHeader;   // bitmap file header
  unsigned char    *bitmapImage;        // bitmap image data
  unsigned int      imageIdx = 0;       // image index counter
  unsigned char     tempRGB;            // swap variable

  // open filename in "read binary" mode
  filePtr = fopen(filename, "rb");
  if (filePtr == NULL)
    return NULL;

  // read the bitmap file header
  fread(&bitmapFileHeader, sizeof(BITMAPFILEHEADER), 1, filePtr);

  // verify that this is a bitmap by checking for the universal bitmap id
  if (bitmapFileHeader.bfType != BITMAP_ID)
  {
    fclose(filePtr);
    return NULL;
  }

  // read the bitmap information header
  fread(bitmapInfoHeader, sizeof(BITMAPINFOHEADER), 1, filePtr);

  // move file pointer to beginning of bitmap data
  fseek(filePtr, bitmapFileHeader.bfOffBits, SEEK_SET);

  // allocate enough memory for the bitmap image data
  bitmapImage = (unsigned char*)malloc(bitmapInfoHeader->biSizeImage);

  // verify memory allocation
  if (!bitmapImage)
  {
    free(bitmapImage);
    fclose(filePtr);
    return NULL;
  }

  // read in the bitmap image data
  fread(bitmapImage, 1, bitmapInfoHeader->biSizeImage, filePtr);

  // make sure bitmap image data was read
  if (bitmapImage == NULL)
  {
    fclose(filePtr);
    return NULL;
  }

  // swap the R and B values to get RGB since the bitmap color format is in BGR
  for (imageIdx = 0; imageIdx < bitmapInfoHeader->biSizeImage; imageIdx+=3)
  {
    tempRGB = bitmapImage[imageIdx];
    bitmapImage[imageIdx] = bitmapImage[imageIdx + 2];
    bitmapImage[imageIdx + 2] = tempRGB;
  }

  // close the file and return the bitmap image data
  fclose(filePtr);
  return bitmapImage;
} // end LoadBitmapFile()


/*****************************************************************************
 LoadBitmapFileWithAlpha

 Loads a bitmap file normally, and then adds an alpha component to use for
 blending
*****************************************************************************/
unsigned char *LoadBitmapFile(char *filename, BITMAPINFOHEADER *bitmapInfoHeader)
{
  unsigned char *bitmapImage = LoadOnlyBitmapFile(filename, bitmapInfoHeader);
  unsigned char *bitmapWithAlpha = (unsigned char *)malloc(bitmapInfoHeader->biSizeImage * 4 / 3);

  if (bitmapImage == NULL || bitmapWithAlpha == NULL)
    return NULL;

  // loop through the bitmap data
  for (unsigned int src = 0, dst = 0; src < bitmapInfoHeader->biSizeImage; src +=3, dst +=4)
  {
    // if the pixel is black, set the alpha to 0. Otherwise, set it to 255.
    if (bitmapImage[src] == 0 && bitmapImage[src+1] == 0 && bitmapImage[src+2] == 0)
      bitmapWithAlpha[dst+3] = 0;
    else
      bitmapWithAlpha[dst+3] = 0xFF;

    // copy pixel data over
    bitmapWithAlpha[dst] = bitmapImage[src];
    bitmapWithAlpha[dst+1] = bitmapImage[src+1];
    bitmapWithAlpha[dst+2] = bitmapImage[src+2];
  }

  free(bitmapImage);

  return bitmapWithAlpha;
} // end LoadBitmapFileWithAlpha()


// LoadBMPTexture()
// desc: loads a texture of the BMP format
texture_t LoadBMPTexture(char *filename)
{
	BITMAPINFOHEADER texInfo;		// BMP header
	texture_t thisTexture;			// the texture


	// store BMP data in texture
	thisTexture.data = LoadBitmapFile(filename, &texInfo);

	// store texture information
	thisTexture.width = texInfo.biWidth;
	thisTexture.height = texInfo.biHeight;
	
	thisTexture.scaledHeight = 0;
	thisTexture.scaledWidth = 0;
	thisTexture.textureType = BMP;

	return thisTexture;
}

//  TGA Loading *********************************************************************
texture_t LoadTGATexture(char *filename)	// Loads A TGA File Into Memory
{
    texture_t texture;
	GLubyte		TGAheader[12]={0,0,2,0,0,0,0,0,0,0,0,0};	// Uncompressed TGA Header
	GLubyte		TGAcompare[12];		// Used To Compare TGA Header
	GLubyte		header[6];		// First 6 Useful Bytes From The Header
	GLuint		bytesPerPixel;	// Holds Number Of Bytes Per Pixel Used In The TGA File
	GLuint		imageSize;	// Used To Store The Image Size When Setting Aside Ram
	GLuint		temp;		// Temporary Variable

	FILE *file = fopen(filename, "rb");		// Open The TGA File

	if(	file==NULL ||			// Does File Even Exist?
		fread(TGAcompare,1,sizeof(TGAcompare),file)!=sizeof(TGAcompare) ||
				// Are There 12 Bytes To Read?
		memcmp(TGAheader,TGAcompare,sizeof(TGAheader))!=0		||
				// Does The Header Match What We Want?
		fread(header,1,sizeof(header),file)!=sizeof(header))// If So Read Next 6 Header Bytes
	{
        if (file == NULL)			// Did The File Even Exist? *Added Jim Strong*
			log_txt("Couldn't load texture1");		// Return False
		else
		{
			fclose(file);		// If Anything Failed, Close The File
			log_txt("Couldn't load texture2");		// Return False
		}
	}


    texture.width  = header[1] * 256 + header[0];		// Determine The TGA Width	(highbyte*256+lowbyte)
	texture.height = header[3] * 256 + header[2];		// Determine The TGA Height	(highbyte*256+lowbyte)

 	if(	texture.width	<=0	||		// Is The Width Less Than Or Equal To Zero
		texture.height	<=0	||	// Is The Height Less Than Or Equal To Zero
		(header[4]!=24 && header[4]!=32))		// Is The TGA 24 or 32 Bit?
	{
		fclose(file);			// If Anything Failed, Close The File
		log_txt("Couldn't load texture3");			// Return False
	}
    texture.bpp	= header[4];		// Grab The TGA's Bits Per Pixel (24 or 32)
	bytesPerPixel	= texture.bpp/8;		// Divide By 8 To Get The Bytes Per Pixel
	imageSize	= texture.width*texture.height*bytesPerPixel;	// Calculate The Memory Required For The TGA Data

    texture.data=(GLubyte *)malloc(imageSize);	// Reserve Memory To Hold The TGA Data

	if(	texture.data==NULL ||		// Does The Storage Memory Exist?
		fread(texture.data, 1, imageSize, file)!=imageSize)
			// Does The Image Size Match The Memory Reserved?
	{
		if(texture.data!=NULL)		// Was Image Data Loaded
			free(texture.data);	// If So, Release The Image Data

		fclose(file);			// Close The File
		log_txt("Couldn't load texture4");			// Return False
	}
    
    for(GLuint i=0; i<int(imageSize); i+=bytesPerPixel)	// Loop Through The Image Data
	{					// Swaps The 1st And 3rd Bytes ('R'ed and 'B'lue)
		temp=texture.data[i];		// Temporarily Store The Value At Image Data 'i'
		texture.data[i] = texture.data[i + 2];
			// Set The 1st Byte To The Value Of The 3rd Byte
		texture.data[i + 2] = temp;// Set The 3rd Byte To The Value In 'temp' (1st Byte Value)
	}
    

	fclose (file);			// Close The File



	return texture;				// Texture Building Went Ok, Return True
}

// LoadTexture()
// desc: loads a texture given the filename
texture_t LoadTexture(char *filename, bool mipmap=true)
{
    log_txt("loading texture: %s",filename);
	texture_t thisTexture[1];
	char *extStr;

	//get extension from filename
	extStr = strrchr(filename, '.');
	extStr++;


	// set the texture type based on extension of filename
	if ((strcmp(extStr, "BMP") == 0) || (strcmp(extStr, "bmp") == 0)){

		thisTexture[0] = LoadBMPTexture(filename);
        thisTexture[0].textureType=BMP;
        thisTexture[0].bpp=32;
    }
	else if ((strcmp(extStr, "TGA") == 0) || (strcmp(extStr, "tga") == 0) ){

		thisTexture[0] = LoadTGATexture(filename);

		thisTexture[0].textureType = TGA;
    }

    log_txt("generating texture");
    // Build A Texture From The Data
    glGenTextures(1, &thisTexture[0].texID);		// Generate OpenGL texture IDs

	glBindTexture(GL_TEXTURE_2D, thisTexture[0].texID);	// Bind Our Texture
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
    
    if(mipmap){
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);	// Linear mpimaps Filtered
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR_MIPMAP_LINEAR);	// Linear mipmaps Filtered
    }
    else{
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);	// Linear Filtered
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);	// Linear Filtered     
    }

    int type=GL_RGB;

    if (thisTexture[0].bpp==32)			// Was The TGA 32 Bits
	{
		type=GL_RGBA;			// If So Set The 'type' To GL_RGBA
	}
    gluBuild2DMipmaps(GL_TEXTURE_2D,type, thisTexture[0].width, thisTexture[0].height, type, GL_UNSIGNED_BYTE, thisTexture[0].data);
    log_txt("Texture information:\n\tbpp: %i\n\twidth: %i\n\theight: %i\n",thisTexture[0].bpp,thisTexture[0].width,thisTexture[0].height);
    log_txt("finished loading texture\n");
	return thisTexture[0];
}

//**********************************************************
//Loads Textures as data, doesn't prepare it for drawing
//**********************************************************


texture_t LoadTexture2(char *filename)
{
    log_txt("loading texture: %s",filename);
	texture_t thisTexture[1];
	char *extStr;

	//get extension from filename
	extStr = strrchr(filename, '.');
	extStr++;


	// set the texture type based on extension of filename
	if ((strcmp(extStr, "BMP") == 0) || (strcmp(extStr, "bmp") == 0)){

		thisTexture[0] = LoadBMPTexture(filename);
        thisTexture[0].textureType=BMP;
        thisTexture[0].bpp=24;
    }
	else if ((strcmp(extStr, "TGA") == 0) || (strcmp(extStr, "tga") == 0) ){

		thisTexture[0] = LoadTGATexture(filename);

		thisTexture[0].textureType = TGA;
    }

    log_txt("Texture information:\n\tbpp: %i\n\twidth: %i\n\theight: %i\n",thisTexture[0].bpp,thisTexture[0].width,thisTexture[0].height);
    log_txt("finished loading texture\n");
	return thisTexture[0];
}

//**********************************************************
//NeHe's font changed to use my textures an alphatesting instead of blending
//**********************************************************


GLvoid glBuildFont(texture_t texture)				// Build Our Font Display List
{
	base=glGenLists(256);			// Creating 256 Display Lists
	glBindTexture(GL_TEXTURE_2D, texture.texID);	// Select Our Font Texture
    //glEnable(GL_BLEND);
    
	for (int loop1=0; loop1<256; loop1++)		// Loop Through All 256 Lists
	{
		float cx=float(loop1%16)/16.0f;	// X Position Of Current Character
		float cy=float(loop1/16)/16.0f;	// Y Position Of Current Character

		glNewList(base+loop1,GL_COMPILE_AND_EXECUTE);	// Start Building A List
			glBegin(GL_QUADS);		// Use A Quad For Each Character
				glTexCoord2f(cx,1.0f-cy-0.0625f);	// Texture Coord (Bottom Left)
				glVertex2d(0,0);	// Vertex Coord (Bottom Left)
				glTexCoord2f(cx+0.0625f,1.0f-cy-0.0625f);// Texture Coord (Bottom Right)
				glVertex2i(16,0);	// Vertex Coord (Bottom Right)
				glTexCoord2f(cx+0.0625f,1.0f-cy-0.001f);// Texture Coord (Top Right)
				glVertex2i(16,16);	// Vertex Coord (Top Right)
				glTexCoord2f(cx,1.0f-cy-0.001f);	// Texture Coord (Top Left)
				glVertex2i(0,16);	// Vertex Coord (Top Left)
			glEnd();			// Done Building Our Quad (Character)
			glTranslated(10,0,0);	// Move To The Right Of The Character
		glEndList();			// Done Building The Display List
	}					// Loop Until All 256 Are Built

     //glDisable(GL_BLEND);
}
		

		
GLvoid glKillFont(GLvoid)				// Delete The Font From Memory
{
	glDeleteLists(base,256);			// Delete All 256 Display Lists
}
		

		
GLvoid glPrint(GLint r, GLint g, GLint b, GLint x, GLint y, const char *fmt, ...)// Where The Printing Happens
{
	char	text[1024];			// Holds Our String
	va_list	ap;				// Pointer To List Of Arguments

	if (fmt == NULL)				// If There's No Text
		return;				// Do Nothing

	va_start(ap, fmt);				// Parses The String For Variables
	    vsprintf(text, fmt, ap);			// And Converts Symbols To Actual Numbers
	va_end(ap);				// Results Are Stored In Text

    y=530-y;

    glColor3f((float)r/255,(float)g/255,(float)b/255);
    //glEnable(GL_BLEND);
    //glBlendFunc(GL_ONE,GL_ONE_MINUS_SRC_COLOR);
    glEnable(GL_TEXTURE_2D);			// Enable Texture Mapping
    glDisable(GL_DEPTH_TEST);				// Disables Depth Testing
    glMatrixMode(GL_PROJECTION);				// Select The Projection Matrix
	glPushMatrix();					// Store The Projection Matrix
    glLoadIdentity();					// Reset The Projection Matrix
	glOrtho(0,800,0,600,-100,100);			// Set Up An Ortho Screen
    glMatrixMode(GL_MODELVIEW);				// Select The Modelview Matrix
	glPushMatrix();					// Store The Modelview Matrix
	glLoadIdentity();					// Reset The Modelview Matrix
    glTranslated(x,y,0);					// Position The Text (0,0 - Bottom Left)
    //glScalef(1.3,1.5,1.0);
    glListBase(base-32);			// Choose The Font Set (0 or 1)
    glCallLists(strlen(text),GL_BYTE,text);		// Write The Text To The Screen
    glMatrixMode(GL_PROJECTION);				// Select The Projection Matrix
	glPopMatrix();						// Restore The Old Projection Matrix
    glMatrixMode(GL_MODELVIEW);				// Select The Modelview Matrix
	glPopMatrix();						// Restore The Old Projection Matrix
    glEnable(GL_DEPTH_TEST);				// Enables Depth Testing
    //glDisable(GL_BLEND);


};

		
GLvoid glPrintSmall(GLint r, GLint g, GLint b, GLint x, GLint y, const char *fmt, ...)// Where The Printing Happens
{
	char	text[1024];			// Holds Our String
	va_list	ap;				// Pointer To List Of Arguments

	if (fmt == NULL)				// If There's No Text
		return;				// Do Nothing

	va_start(ap, fmt);				// Parses The String For Variables
	    vsprintf(text, fmt, ap);			// And Converts Symbols To Actual Numbers
	va_end(ap);				// Results Are Stored In Text

    y=530-y;

    glColor3f((float)r/255,(float)g/255,(float)b/255);
    //glEnable(GL_BLEND);
    //glBlendFunc(GL_ONE,GL_ONE_MINUS_SRC_COLOR);
    glEnable(GL_TEXTURE_2D);			// Enable Texture Mapping
    glDisable(GL_DEPTH_TEST);				// Disables Depth Testing
    glMatrixMode(GL_PROJECTION);				// Select The Projection Matrix
	glPushMatrix();					// Store The Projection Matrix
    glLoadIdentity();					// Reset The Projection Matrix
	glOrtho(0,800,0,600,-100,100);			// Set Up An Ortho Screen
    glMatrixMode(GL_MODELVIEW);				// Select The Modelview Matrix
	glPushMatrix();					// Store The Modelview Matrix
	glLoadIdentity();					// Reset The Modelview Matrix
    glTranslated(x,y,0);					// Position The Text (0,0 - Bottom Left)
    glScalef(1.3,1.5,1.0);
    glListBase(base-32);			// Choose The Font Set (0 or 1)
    glCallLists(strlen(text),GL_BYTE,text);		// Write The Text To The Screen
    glMatrixMode(GL_PROJECTION);				// Select The Projection Matrix
	glPopMatrix();						// Restore The Old Projection Matrix
    glMatrixMode(GL_MODELVIEW);				// Select The Modelview Matrix
	glPopMatrix();						// Restore The Old Projection Matrix
    glEnable(GL_DEPTH_TEST);				// Enables Depth Testing
    //glDisable(GL_BLEND);


};


//**********************************************************
//Texture Handling
//**********************************************************


void glBindTex(texture_t texture,int ARBnr=0){
     if(texture.textureType==BMP){
                                  glEnable(GL_BLEND);
                                  glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
                                  glEnable(GL_ALPHA_TEST);
                                  glAlphaFunc(GL_GREATER,0);
                                  }
     if(multitex){
                  switch(ARBnr){
                                case 0: glActiveTextureARB(GL_TEXTURE0_ARB);break;
                                case 1: glActiveTextureARB(GL_TEXTURE1_ARB);break;
                                case 2: glActiveTextureARB(GL_TEXTURE2_ARB);break;
                                case 3: glActiveTextureARB(GL_TEXTURE3_ARB);break;
                  }
                  glEnable(GL_TEXTURE_2D);
                  glBindTexture(GL_TEXTURE_2D,texture.texID);
     }
     else if(ARBnr==0)
                  glBindTexture(GL_TEXTURE_2D,texture.texID);


};

void glUnbindAll(){
     
     glDisable(GL_BLEND);
                                  
     glDisable(GL_ALPHA_TEST);
                                  
    
     if(multitex){
     glActiveTextureARB(GL_TEXTURE0_ARB);
     glDisable(GL_TEXTURE_2D);
     glActiveTextureARB(GL_TEXTURE1_ARB);
     glDisable(GL_TEXTURE_2D);
     glActiveTextureARB(GL_TEXTURE2_ARB);
     glDisable(GL_TEXTURE_2D);
     glActiveTextureARB(GL_TEXTURE3_ARB);
     glDisable(GL_TEXTURE_2D);
     };
}

void glTexCoordARB(float s, float t,int ARBnr=0){
     if(multitex){
                switch(ARBnr){
                                case 0: glMultiTexCoord2fARB(GL_TEXTURE0_ARB,s,t);break;
                                case 1: glMultiTexCoord2fARB(GL_TEXTURE1_ARB,s,t);break;
                                case 2: glMultiTexCoord2fARB(GL_TEXTURE2_ARB,s,t);break;
                                case 3: glMultiTexCoord2fARB(GL_TEXTURE3_ARB,s,t);break;
                  }
     }
     else if(ARBnr==0)
         glTexCoord2f(s,t);
}

//**********************************************************
// extensions for Drawing
//**********************************************************

#define glTexVert(s,t,x,y,z) glTexCoordARB(s,t); glVertex3f(x,y,z)



#endif
